package sw

import (
	"context"
	"sync"
	"time"
)

// Window 窗口
type Window interface {
	Start() time.Time
	Count() int64
	AddCount(n int64)
	Reset(s time.Time, c int64)
	Sync(ctx context.Context, now time.Time)
}

type StopFunc func()

type NewWindow func() (Window, StopFunc)

type Limiter struct {
	size  time.Duration
	limit int64
	mu    sync.Mutex
	curr  Window
	prev  Window
}

func NewLimiter(size time.Duration, limit int64, newWindow NewWindow) (*Limiter, StopFunc) {
	currWin, currStop := newWindow()
	prevWin, _ := NewLocalWindow()
	lim := &Limiter{
		size:  size,
		limit: limit,
		curr:  currWin,
		prev:  prevWin,
	}
	return lim, currStop
}

func (lim *Limiter) Size() time.Duration {
	return lim.size
}

func (lim *Limiter) Limit() int64 {
	lim.mu.Lock()
	defer lim.mu.Unlock()
	return lim.limit
}

func (lim *Limiter) SetLimit(newLimit int64) {
	lim.mu.Lock()
	defer lim.mu.Unlock()
	lim.limit = newLimit
}

func (lim *Limiter) Allow(ctx context.Context) bool {
	return lim.AllowN(ctx, time.Now(), 1)
}

func (lim *Limiter) AllowN(ctx context.Context, now time.Time, n int64) bool {
	lim.mu.Lock()
	defer lim.mu.Unlock()
	lim.advance(now)
	elapsed := now.Sub(lim.curr.Start())
	weight := float64(lim.size-elapsed) / float64(lim.size)
	count := int64(weight*float64(lim.prev.Count())) + lim.curr.Count()
	defer lim.curr.Sync(ctx, now)
	if count+n > lim.limit {
		return false
	}
	lim.curr.AddCount(n)
	return true
}

// advance更新由于时间的推移而产生的当前/以前的Windows。
func (lim *Limiter) advance(now time.Time) {
	//计算当前窗口的起始边界。
	newCurrStart := now.Truncate(lim.size)
	diffSize := newCurrStart.Sub(lim.curr.Start()) / lim.size
	if diffSize >= 1 {
		//当前窗口至少比预期窗口小一个窗口大小。
		newPrevCount := int64(0)
		if diffSize == 1 {
			//新的previous-window将与旧的current-window重叠，因此它继承了计数。
			//注意这里的计数可能不准确，因为它只是一个 当前窗口计数的快照，其本身趋向于由于同步行为的异步性质而不准确。
			newPrevCount = lim.curr.Count()
		}
		lim.prev.Reset(newCurrStart.Add(-lim.size), newPrevCount)
		//新的当前窗口计数总是为零。
		lim.curr.Reset(newCurrStart, 0)
	}
}
